open GrailAbsyn;
open Bytecode;
open Classdecl;
open Jvmtype;
open ToyGrailAbsyn;

exception gdfError of string;

type ident = string;

fun fst (x,y) = x;
fun snd (x,y) = y;

fun member x [] = false
  | member x (h::t) = x=h orelse member x t;

(*val mapSnd = map snd;*)

fun pr x = TextIO.print (x^" ");
fun prl x = TextIO.print (x^"\n");

fun pr x = ();
fun prl x = ();


(*************************************)
(*           Conversions             *)
(*************************************)

fun tyToJty INTty = Tint
  | tyToJty BOOLEANty = Tboolean
  | tyToJty FLOATty = Tfloat
  | tyToJty (REFty x) = Tclass (qnameToJClass x)
  | tyToJty (ARRAYty ty) = Tarray (tyToJty ty)
  | tyToJty _ = raise gdfError "Not implemented yet"

and rtyToJty (SOME t) = SOME (tyToJty t)
  | rtyToJty NONE = NONE

and qnameToJClass name = (pr ("# class " ^ name ^ "\n");
    if name ="" then class{pkgs = [], name = ""} else 
    let val l = String.tokens (fn c => (c = #".")) name
	val n = List.length l
	val _ = app (fn x => (pr (x ^ "\n"))) l
	val _ = pr ((Int.toString n) ^"\n")
	val pkgs = List.take (l,n-1)
	val name' = List.nth (l,n-1)
	val _ = prl "@ Classname" 
	val _ = pr "Packages: [" 
	val _ = app pr pkgs;
	val _ = prl "]"
	val _ = prl ("Class:" ^ name')
    in class {pkgs = pkgs, name = name'}
    end)

and mdescToMethodref (MDESC(ty, name, argtypes)) = 
    let val _ = prl ("## method " ^ name );
	val l = String.tokens (fn c => (c = #".")) name;
	val n = List.length l;
	val msig = (map tyToJty argtypes, rtyToJty ty);
	val name = List.nth(l,n-1);
	val (pkgs, cname) = 
	    if n=1 then
		(TextIO.print ("WARNING: suspect method descriptor (" ^ name ^ ")\n");
		 ([],"")
		 )
	    else (List.take(l,n-2), List.nth(l,n-2));
	val _ = prl "@ Methodref" ;
	val _ = pr "Packages: [" ;
	val _ = app pr pkgs;;
	val _ = prl "]";
	val _ = prl ("Class: " ^ cname);
	val _ = prl ("Name: " ^ name);
    in 
	{name=name, msig=msig, class = Jvmtype.class{pkgs=pkgs, name=cname}}
    end


and fdescToFieldRef (FDESC(fty,name)) =
  let val l = String.tokens (fn c => (c = #".")) name
	val n = List.length l
	val ty = tyToJty fty
	val name = List.nth(l,n-1)
	val (pkgs, cname) = 
	    if n=1 then 
		(TextIO.print ("WARNING: suspect field descriptor (" ^ name ^ ")\n");
		 ([],"")
		 )
	    else (List.take(l,n-2), List.nth(l,n-2))
	(* val _ =BasicIO.print ("cname = "^cname^", field = "^ name ^"\n") *)
  in 
      {name=name, ty=ty, class = class{pkgs=pkgs, name=cname}}
  end;

(*************************************)
(*    Local variables and labels     *)
(*************************************)

fun lookup  v (ref l) = let fun
    lookup' v [] = NONE
  | lookup' v ((w,p)::tail) = if v=w then SOME p else lookup' v tail
in 
    lookup' v l
end
;

fun find v h = case lookup v h of 
    NONE => raise gdfError ("Can't find variable/function " ^ v)
  | SOME p => p
;


(* Local Variables and labels for function declarations and calls *)


local
    (* locals *)
    val localsref = ref Localvar.freshLocals : Localvar.locals ref

    fun newLocal1 () = 
        let val (locals, loc) = Localvar.nextVar1 (!localsref)
        in localsref := locals; loc 
	end

    val registers: (ident * (Ty * Localvar.index)) list ref
	= ref []  (* (name, (type, varpos)) *)

    (* labels *)
    val nextlab = ref Label.freshLabels : Label.labels ref
    val (_,cs) = Label.newLabel(!nextlab)   (* Horrible *)
    val (_,ce) = Label.newLabel(!nextlab)   (* Horrible *)
    val funLabels: (ident * (RTy * (Ty*string) list * Label.label)) list ref 
	= ref [];

in      
    val codeStart = ref cs;
    val codeEnd = ref cs;  (* Both of these are initial values which will never be used *)

    fun resetLocals () = (localsref := Localvar.freshLocals; registers := [])
    fun numLocals() = List.length (!registers)

    fun lookupVar v = lookup v registers
    fun findVar v = find v registers

    fun declareVar v ty =
	case lookupVar v of 
	    SOME (t1,_) => 
		if (t1<>ty) then
		    raise gdfError ( "Trying to reuse variable " 
				      ^ v ^ " (of type " ^ tyToString t1
				      ^ ") with new type " ^ tyToString ty)
		else ()
	  | NONE =>
		let 
		    val _ = pr ("Declaring " ^ v ^ "\n") 
		    val loc = newLocal1() 
		in 
		    registers := (v, (ty, loc))::(!registers) 
		end
	    
    fun dumpLocals () =
	let 
	    val l = !registers
	    fun dl [] n = ()
	      | dl ((name,_)::t) n = 
		(BasicIO.print ((Int.toString n) ^ ": " ^ name ^"\n"); 
		 dl t (n-1))
	in dl l (List.length l - 1)
	end

    fun getLocals () =
	let 
	    fun get' [] = []
	      | get'  ((id,(ty,loc))::t) = 
		{from=(!codeStart), thru=(!codeEnd), name=id, ty=tyToJty(ty), index=loc}::get'(t)
	in
	    LOCALVAR(get' (!registers))
	end



    (*labels*)

    fun newLabel () = (* We need this to be globally visible so that 
		         we can generate new labels for "if ..." *)
        let val (labels, lab) = Label.newLabel (!nextlab)
        in nextlab := labels; lab
	end

    fun resetLabels () = 
	(nextlab := Label.freshLabels; 
	 codeStart := newLabel();
	 codeEnd := newLabel();
	 funLabels := [])
	
    fun lookupFun f = lookup f funLabels;
    fun findFun f = find f funLabels;
	
    fun declareFun name ty params =
	case lookupFun name of 
	    SOME _ => raise gdfError ( "Trying to redeclare function " ^ name )
	  | NONE => let val _ = pr("Declaring function " ^ name ^ "\n") 
			val loc = newLabel() 
		    in 
			funLabels := (name, (ty, params, loc))::(!funLabels) 
		    end

    fun getFunInfo () =
	let 
	    fun locOfParam (_, name) = 
		let val (_, loc) = findVar name
		in 
		    Localvar.toInt loc
		end

	    fun get' [] = []
	      | get'  ((ident, (rty, params, loc))::t) = 
		(ident, map locOfParam params)::(get' t)
		

(*		
	    fun p (l,i,m) = 
		let 
		    val pr = TextIO.print 
		in
		    pr (Label.toString l);
		    pr ( " -> ");
		    pr i;
		    pr ": ";
		    app (fn x => pr (Int.toString(x) ^ " ")) m;
		    pr "\n"
		end

	    val _ = app p (get'(!funLabels))
*)

	    fun stringToBytes s =
		(String.size s) :: (map Char.ord (String.explode s))

	    fun makeByteList [] = []
	      | makeByteList l = 
		let 
		    fun mk [] acc = acc
		      | mk ((i,m)::t) acc = 
			mk t ((stringToBytes i) @ ((length m)::m) @ acc)
		in
		    mk l []
		end

	    val labelInfo = makeByteList (get' (!funLabels))
(*	    val _ = prIntList labelInfo*)


	(* now encode this as as Word8Vector *)
	    val info = Word8Vector.fromList (map Word8.fromInt labelInfo)

	in	 
	    ATTR {attr="uk.ac.ed.dcs.mrg.GrailFunctionInfo", 
		  info=info}
	end
end
;




(*************************************)
(*           Typechecking            *)  
(*************************************)

fun typeOfVar v =
    case lookupVar v of
	NONE => raise gdfError ("Can't infer type of variable " ^ v)
      | SOME (t,_) => t
 
fun typeOfValue (VARval v)  = typeOfVar v
  | typeOfValue (INTval _)  = INTty
  | typeOfValue (FLOATval _) = FLOATty
  | typeOfValue (STRINGval _) = REFty "java.lang.String"
  | typeOfValue (NULLval t) = REFty t

fun typeOfPrimop (VALop v) = SOME(typeOfValue v)
  | typeOfPrimop (BINop(oper,v,w)) = 
    let val t1 = typeOfValue v
	val t2 = typeOfValue w
    in 
	if t1=INTty andalso t2=INTty 
	    then SOME(INTty)
	else if t1=FLOATty andalso t2=FLOATty
	    then SOME(FLOATty)
	else
	    raise gdfError ("Arithmetic operator applied to non-numeric/mixed types")
    end
  | typeOfPrimop (NEWop (MDESC(_,s,_),_) ) =              SOME(REFty s)
  | typeOfPrimop (CHECKCASTop (s,_)) =                    SOME(REFty s)
  | typeOfPrimop (INSTANCEop  _)     =                    SOME(INTty)
  | typeOfPrimop (GETFIELDop (_,FDESC (ty,_))) =          SOME(ty)
  | typeOfPrimop (GETSTATICop (FDESC (ty,_))) =           SOME(ty)
  | typeOfPrimop (INVOKEVIRTUALop (_,MDESC(rty,_,_),_)) = rty
  | typeOfPrimop (INVOKESPECIALop (_,MDESC(rty,_,_),_)) = rty
  | typeOfPrimop (INVOKESTATICop (MDESC(rty,_,_),_))    = rty
  | typeOfPrimop (PUTFIELDop _)  =                        NONE
  | typeOfPrimop (PUTSTATICop _) =                        NONE
  | typeOfPrimop (MAKEop (a,_))  = raise gdfError "The make operation is not allowed yet"
                                  (* SOME(ARRAYty(typeOfValue(a))) *)
  | typeOfPrimop (GETop (aa,_))  = 
    (
     case typeOfValue(aa) of 
	 ARRAYty(ty) => SOME(ty)
       | _ => (raise gdfError ("First argument of get must be of array type"))
	      )
  | typeOfPrimop (SETop (aa,_,_)) = NONE
  | typeOfPrimop (LENGTHop _)     = SOME(INTty)
  | typeOfPrimop (EMPTYop(v,t))   = SOME(ARRAYty(t)) 
  | typeOfPrimop (FTOIop _)       = SOME(INTty)
  | typeOfPrimop (ITOFop _)       = SOME(FLOATty)

fun tysEqual x y =
    x = y
    orelse (x = INTty andalso y = BOOLEANty)
    orelse (x = BOOLEANty andalso y = INTty)
and tysUnequal x y = not (tysEqual x y)
and rtysEqual (SOME x) (SOME y) = tysEqual x y
  | rtysEqual NONE NONE = true
  | rtysEqual _ _ = false
and rtysUnequal x y = not (rtysEqual x y)

fun typeCheckFunCall fname params = 
    let 
	val (_,formalParams,_) = findFun fname
	val formalNames = map snd formalParams
    in 
	if params <> formalNames then
	    raise gdfError ("Function " ^ fname ^ " is called with arguments " ^ 
			      strListToString(params) ^
			      ", but requires arguments " ^ strListToString(formalNames))

(*	else if map typeOfValue params <> map fst formalParams then
	    raise gdfError ("INTERNAL ERROR: type mismatch in function call") *)
(* Doesn't work, but shouldn't happen *)
	else ()
    end
	


fun typeOfPrimRes (OPres p) = typeOfPrimop p
  | typeOfPrimRes VOIDres = NONE
  | typeOfPrimRes (FUNres (fname, params)) = 
    (
(* should check that  params are in scope & have been declared *)
     typeCheckFunCall fname params;
     case (lookupFun fname) of 
	 NONE => (raise gdfError ("Type error: can't find function " ^ fname))
       | SOME(t, n, p) => t 
    )

fun numeric INTty = true
  | numeric FLOATty = true
  | numeric _ = false
    
fun typeOfResult (PRIMres p) = typeOfPrimRes p
  | typeOfResult (CHOICEres (v,tst,w, p1,p2)) = 
    let val t = typeOfValue v
	val t' = typeOfValue w
	val () = if tysUnequal t t' 
		     then raise gdfError ("Test arguments have different types (" ^ 
					  tyToString t ^ " and " ^ tyToString t' ^")")
		 else ()
	val () = case tst of 
	    EQtest => ()
	  | NEtest => () 
	  | _ => if not (numeric t)
		     then 
			 raise gdfError "Numerical test applied to non-numerical values"
		 else ()
	val t1 = typeOfPrimRes p1
	val t2 = typeOfPrimRes p2			  

    in 
	if rtysUnequal t1 t2 then raise gdfError ("Test branches have different types") 
	else t1
    end

fun typeCheckLetDec fname (VOIDdec p) =
    let val ty = typeOfPrimop p in
    if (ty <> NONE) then 
	raise gdfError ("In function " ^ fname ^ ", attempted to assign value of type " ^
			  rtyToString(ty) ^ " to ()")
    else () 
    end
  | typeCheckLetDec fname (VALdec (v, p)) = 
    let 
	val lty = typeOfVar v
	val ty = typeOfPrimop p (* Shouldn't be NONE by this stage *)
    in
	if rtysUnequal (SOME lty) ty then 
	raise gdfError ("In function " ^ fname ^ ", attempted to assign value of type " ^
			  rtyToString(ty) ^ " to variable " ^ 
			  v ^ " of type " ^ rtyToString(SOME lty))
    else () 
    end


fun typeCheckReturnType expected fname actual =
    if actual <> expected then
      raise gdfError("Function " ^ fname ^ " should return " ^ rtyToString(expected) ^
		       ", but actually returns " ^ rtyToString(actual))
      else ()

fun typeCheckFunBody rtype fname (FUNbody(lets, r)) =
       (app (typeCheckLetDec fname) lets;        
	     typeCheckReturnType rtype fname (typeOfResult r) )

fun typeCheckFunDec rtype (FDEC(fname,_,fbody)) = typeCheckFunBody rtype fname fbody

fun typeCheckMbody rtype (MBODY(_, fundecs, result)) =  
    let 
	val actualrtype = typeOfResult result
    in 
	if  actualrtype <> rtype
	    then raise gdfError ("Method should return " ^ rtyToString rtype ^ 
		", but actually returns " ^ rtyToString actualrtype)
	else app (typeCheckFunDec rtype) fundecs
    end
;


(***********************************************)
(* Declarations & type inference for variables *)  
(***********************************************)


(* make sure that all variables referred to in a primop 
   have been declared in the current function *)

fun checkDec fname p vars = 
    let 
	fun chkvar v = 
	    if member v vars then () 
	    else raise gdfError ("Unknown variable " ^ v ^ " in " ^ fname )

	fun chkval w =
	    case w of (VARval v) => chkvar v
	  | _ => ()

	fun chkvlist [] = ()
	  | chkvlist (h::t) = (chkval h; chkvlist t)

	fun chk p = case p of
	    VALop w => chkval w
	  | BINop(_,v,w) => (chkval v; chkval w)
	  | NEWop(MDESC(_,s,_),vs) => (chkvlist vs)
	  | CHECKCASTop (_,v) => chkvar v
	  | INSTANCEop  (_,v) => chkvar v
	  | INVOKEVIRTUALop (v,_,l) => (chkvar v; chkvlist l)
	  | INVOKESPECIALop (v,_,l) => (chkvar v; chkvlist l)
	  | INVOKESTATICop (_,l) => chkvlist l
	  | GETFIELDop  (v,_) => chkvar v
	  | PUTFIELDop (v,_,w) => (chkvar v; chkval w)
	  | GETSTATICop _ => ()
	  | PUTSTATICop (_,v) => chkval v
          | MAKEop (v,i) => (chkval v; chkval i)
          | GETop (a,i) => (chkval a; chkval i)
          | SETop (a,i,v) => (chkval a; chkval i; chkval v)
          | LENGTHop a => chkval a
          | EMPTYop (v, _) => chkval v
          | ITOFop i => chkval i
          | FTOIop f => chkval f
    in
	chk p
    end
;


(******* ATTENTION: this doesn't work properly  ************)
(* 
  class w {
	method public static int f (int x, int y) =
	let
		fun f (int x, int t) = if x < 0 then x else t
	in
	        f (x,t)
	end
  }

  will compile successfully,  but t is never declared 
*)
(***********************************************************)


(* You should only be able to write f(x) if x was supplied
   as an argument to the current function,  or was previously
   declared in the current function using "val" )
*)


fun declareParams params = 
    app (fn (t,v) => declareVar v t) params;

fun declareLetDecs _ [] _ = () 
  | declareLetDecs fname (h::t) vars = 
    (
     case h of VOIDdec(p) =>
	 (
	  checkDec fname p vars;
	  if typeOfPrimop p = NONE 
	      then declareLetDecs fname t vars
	  else raise gdfError ("Non-unit value assigned to () in " ^ fname )
	 )
   | VALdec (v,p) =>
	 (
	  checkDec fname p vars;
	  case typeOfPrimop p of
	      NONE => raise gdfError 
		  ("You're trying to assign a void value to the variable " ^ v ^ " in " ^ fname)
	    | SOME ty =>  (declareVar v ty; declareLetDecs fname t (v::vars))
	 )
)
	    
	    
	    
fun declareFunDecs fundecs = 
    app (fn FDEC (fname, params, FUNbody (lets, _)) =>
	 ( declareParams params; declareLetDecs fname lets (map snd params) ))
    fundecs 
;

fun declareMdefVars cname mname flags params body =
    let
        val params = 
	    if member ACCstatic flags then params 
	    else (REFty(cname), "this")::params 
    (* instance method - fake declaration of "this" *)

    in
        declareParams params;
	let val MBODY(letdecs, fundecs,_) = body in
	    (declareLetDecs mname letdecs (map snd params); 
	     declareFunDecs fundecs ) 
	end
    end

fun declareFuns methty mbody =
    let val MBODY(_, fundecs,_) = mbody in 
	app (fn FDEC(name, params, _) 
	     => declareFun name methty params) 
	fundecs
    end
;


(*************************************)
(*       Bytecode production         *)  
(*************************************)

fun loadVar v A = (
   case lookupVar v  of   (* should really use findVar here *)
     NONE => raise gdfError ("Eh? can't find variable " ^ v)
   | SOME (t,l) => 
      (	case t of 
         INTty => (Jiload l) :: A
       | BOOLEANty => (Jiload l) :: A
       | FLOATty => (Jfload l) :: A
       | REFty _ => (Jaload l) :: A
       | ARRAYty _ => (Jaload l) :: A
       | _ => raise gdfError "Not implemented yet"
      )
)
;

fun storeVar v A = (
   case lookupVar v of   (* should really use findVar here *)
     NONE => raise gdfError ("Eh? can't find variable " ^ v)
   | SOME (t,l) => 
      (	case t of 
	    INTty => (Jistore l) :: A
          | BOOLEANty => (Jistore l) :: A
          | FLOATty => (Jfstore l) :: A
	  | REFty _ => (Jastore l) :: A
	  | ARRAYty _ => (Jastore l) :: A
	  | _ => raise gdfError "Not implemented yet"
      )
)
;

fun loadVal (VARval v) A = loadVar v A
  | loadVal (INTval i) A = (Jiconst (Int32.fromInt i)) :: A
  | loadVal (FLOATval r) A = (Jfconst (Real32.fromReal r)) :: A
  | loadVal (STRINGval s) A = 
    (case (String.fromCString s) of  (* quotes should have been removed by now *)
	SOME(x) => (Jsconst x) :: A
      | NONE => raise gdfError ("Error in string literal"))
  | loadVal (NULLval _) A = (Jaconst_null) :: A
;

fun loadArgs [] A = A 
  | loadArgs (h::t) A = loadVar h (loadArgs t A)
;

fun loadArgVals [] A = A
  | loadArgVals (h::t) A = loadVal h (loadArgVals t A)
;

fun storeArgs l A = 
let
    fun storeArgs' [] A = A
      | storeArgs' ((ty,name)::t) A = storeVar name (storeArgs' t A)
in
    storeArgs' (List.rev l) A
end
;


fun compilePrimOp (VALop v) A = loadVal v A
  | compilePrimOp (BINop (ADDop, a, b)) A = 
    if typeOfValue a = INTty then
	loadVal a (loadVal b (Jiadd :: A))
    else
	loadVal a (loadVal b (Jfadd :: A))
  | compilePrimOp (BINop (SUBop, a, b)) A = 
    if typeOfValue a = INTty then
	loadVal a (loadVal b (Jisub :: A))
    else
	loadVal a (loadVal b (Jfsub :: A))
  | compilePrimOp (BINop (MULop, a, b)) A = 
    if typeOfValue a = INTty then
	loadVal a (loadVal b (Jimul :: A))
    else
	loadVal a (loadVal b (Jfmul :: A))
  | compilePrimOp (BINop (DIVop, a, b)) A = 
    if typeOfValue a = INTty then
	loadVal a (loadVal b (Jidiv :: A))
    else
	loadVal a (loadVal b (Jfdiv :: A))
  | compilePrimOp (BINop (MODop, a, b)) A = 
    if typeOfValue a = INTty then
	loadVal a (loadVal b (Jirem :: A))
    else
	loadVal a (loadVal b (Jfrem :: A))
  | compilePrimOp (NEWop(MDESC(r,s,tl), vs)) A = 
    (Jnew (qnameToJClass s)) :: Jdup :: (loadArgVals vs
     (Jinvokespecial (mdescToMethodref (MDESC(r, s ^ ".<init>", tl))) :: A))
  | compilePrimOp (CHECKCASTop (s,v)) A = 
	loadVar v  ((Jcheckcast(CLASS (qnameToJClass s))) :: A)
  | compilePrimOp (INSTANCEop  (s,v)) A = 
        loadVar v (Jinstanceof(CLASS (qnameToJClass s)) :: A)
  | compilePrimOp (INVOKESTATICop (md, args)) A = 
        loadArgVals(args) (Jinvokestatic (mdescToMethodref md) :: A)
  | compilePrimOp (INVOKEVIRTUALop (inst, md, args)) A = 
    let val (_,loc) = findVar inst in 
        (Jaload loc) :: (loadArgVals args (Jinvokevirtual (mdescToMethodref md) :: A))
    end
  | compilePrimOp (INVOKESPECIALop (inst, md, args)) A = 
    let val (_,loc) = findVar inst in 
        (Jaload loc) :: (loadArgVals args (Jinvokespecial (mdescToMethodref md) :: A))
    end
  | compilePrimOp (GETFIELDop (inst ,fd)) A =
     let val (_,loc) = findVar inst in 
	 (Jaload loc) :: (Jgetfield (fdescToFieldRef fd) :: A)
     end
  | compilePrimOp (PUTFIELDop (inst, fd, input)) A = 
     let val (_,loc) = findVar inst in
       (Jaload loc) :: (loadVal input (Jputfield (fdescToFieldRef fd) :: A))
     end
  | compilePrimOp (GETSTATICop (fd)) A = (Jgetstatic (fdescToFieldRef fd)) :: A
  | compilePrimOp (PUTSTATICop (fd, input)) A = 
       loadVal input (Jputstatic (fdescToFieldRef fd) :: A)
  | compilePrimOp (GETop(aa,i)) A = 
      (
       case typeOfValue aa of 
	   ARRAYty(INTty)     => loadVal aa (loadVal i (Jiaload :: A))
	 | ARRAYty(FLOATty)   => loadVal aa (loadVal i (Jfaload :: A))
	 | ARRAYty(REFty _)   => loadVal aa (loadVal i (Jaaload :: A))
	 | ARRAYty(ARRAYty _) => loadVal aa (loadVal i (Jaaload :: A))
	 | _ => (raise gdfError("First argument of get must be an array type"))
       )
  | compilePrimOp (SETop(aa,i,v)) A = (* Changed to return no value = no dup *)
      (
       case typeOfValue aa of
	   ARRAYty(INTty)     => loadVal aa (loadVal i (loadVal v (Jiastore :: A)))
	 | ARRAYty(FLOATty)   => loadVal aa (loadVal i (loadVal v (Jfastore :: A)))
	 | ARRAYty(REFty _)   => loadVal aa (loadVal i (loadVal v (Jaastore :: A)))
	 | ARRAYty(ARRAYty _) => loadVal aa (loadVal i (loadVal v (Jaastore :: A)))
	 | _ => (raise gdfError("First argument of set must be an array type"))
       )
  | compilePrimOp (LENGTHop(aa)) A = loadVal aa (Jarraylength :: A)
  | compilePrimOp (EMPTYop(i,ty)) A = 
         loadVal i ( (Jnewarray { elem = tyToJty ty, dim = 1 }) :: A)
  | compilePrimOp (MAKEop(v,i)) A =
       loadVal v  ((Jnewarray { elem = tyToJty (typeOfValue v), dim = 1 }) :: A)
  | compilePrimOp (FTOIop f) A = loadVal f (Jf2i :: A)
  | compilePrimOp (ITOFop i) A = loadVal i (Ji2f :: A)
   
;

fun compileLetDec (VALdec (v,p)) A = compilePrimOp p (storeVar v A)
  | compileLetDec (VOIDdec p) A = compilePrimOp p A
;

fun compileLetDecs letdecs A =
    let fun comp [] A = A
	  | comp (h::t) A = compileLetDec h (comp t A)
    in
	comp letdecs A
end
;


fun compilePrimRes (OPres p) A = compilePrimOp p (Jreturn :: A)
  | compilePrimRes VOIDres  A = Jreturn :: A
  | compilePrimRes (FUNres (fname, params)) A = 
    let val (_,_,dest) = findFun fname
    in
	(Jgoto dest) :: A
    end
;

fun compileTest INTty EQtest dest A = (Jif_icmpeq dest) :: A         (* EQtest *)
  | compileTest BOOLEANty EQtest dest A = (Jif_icmpeq dest) :: A
  | compileTest FLOATty EQtest dest A = 
    Jfcmpl :: (Jifeq dest) :: A
    (* fcmpl & gcmpl only differ on NaN *)
  | compileTest _ EQtest dest A = (Jif_acmpeq dest) :: A

  | compileTest INTty NEtest dest A = (Jif_icmpne dest) :: A         (* NEtest *)
  | compileTest BOOLEANty NEtest dest A = (Jif_icmpne dest) :: A
  | compileTest FLOATty NEtest dest A = 
    Jfcmpl :: (Jifne dest) :: A
    (* fcmpl & gcmpl only differ on NaN *)
  | compileTest _ NEtest dest A = (Jif_acmpne dest) :: A

  | compileTest INTty Ltest dest A =   (Jif_icmplt dest) :: A        (* Ltest *)
  | compileTest FLOATty Ltest dest A = Jfcmpl :: (Jiflt dest) :: A
  | compileTest _ Ltest _ _ = 
    raise gdfError("< may only be applied to numerical types")

  | compileTest INTty LEtest dest A =   (Jif_icmple dest) :: A       (* LEtest *)
  | compileTest FLOATty LEtest dest A = Jfcmpl :: (Jifle dest) :: A
  | compileTest _ LEtest _ _ = 
    raise gdfError("<= may only be applied to numerical types")

  | compileTest INTty Gtest dest A =   (Jif_icmpgt dest) :: A        (* Gtest *)
  | compileTest FLOATty Gtest dest A = Jfcmpl :: (Jifgt dest) :: A
  | compileTest _ Gtest _ _ = 
    raise gdfError("> may only be applied to numerical types")

  | compileTest INTty GEtest dest A =   (Jif_icmpge dest) :: A       (* GEtest *)
  | compileTest FLOATty GEtest dest A = Jfcmpl :: (Jifge dest) :: A
  | compileTest _ GEtest _ _ = 
    raise gdfError(">= may only be applied to numerical types")
;

fun compileResult (PRIMres p) A = compilePrimRes p A 
       | compileResult (CHOICEres (v, test, w, p1, p2)) A =
	     let val dest1 = newLabel()
	     in
		 loadVal v (
	          loadVal w (
                   compileTest (typeOfValue v) test dest1 ( 
		    compilePrimRes p2 ((Jlabel dest1) :: (compilePrimRes p1 A)))))
	     end
;
			   
fun compileFunBody (FUNbody(letdecs, result)) A 
		       = compileLetDecs letdecs (compileResult result A)

(* Now we're compiling a fun declaration.  At this point, 
   the name of the function and its local variables have already
   been declared *)

fun compileFunDec (FDEC(fname, args, body)) A =
    let 
	val (_,_,entry) = findFun fname
    in
	Jlabel(entry) :: (compileFunBody body A)
    end
;
    
fun compileFunDecs l A = 
    let fun 
	compileFuns [] A = A
      | compileFuns (h::t) A = compileFunDec h (compileFuns t A)
    in compileFuns l A end
;

fun compileMethodBody (MBODY(letdecs,fundecs,result)) = 
	Jlabel(!codeStart) :: (
         compileLetDecs letdecs (
	  compileResult result (
	   compileFunDecs fundecs [Jlabel(!codeEnd)])))
;
		       
fun compile_mdef cname (MDEF(flags, rettype, mname, params, mbody)) = 
    let 
	val () = resetLocals ();
	val () = resetLabels ();
	val () = declareMdefVars cname mname flags params mbody
	val () = declareFuns rettype mbody
	val () = typeCheckMbody rettype mbody

	val msig =(map (fn (t,v) => tyToJty t) params, rtyToJty rettype) 
	val code = compileMethodBody mbody
	val attrs = [CODE{stack=Stackdepth.maxdepth code [], 
			  locals=numLocals(),
			  code=code, hdls=[], 
			  attrs=[getLocals(), getFunInfo()]}]
(*	val _ = TextIO.print (Int.toString(numLocals ()) ^ " local variables\n");*)
(*	val _ = dumpLocals();*)

    in 
	{flags=flags, name=mname, msig=msig, attrs=attrs}  
    end;


fun add_initialiser l = 
(* Add a default initialsier if the user hasn't defined one.
   This assumes that the superclass is Object.  Syntax of 
   initialisers is complex,  so we should HAVE ANOTHER LOOK AT THIS *)
let

    val defInit: method_ref = {class=class{pkgs=["java","lang"], name="Object"},
			       name="<init>",
			       msig=([]: jtype list, NONE: jtype option)}
			       
    val init = {flags = [ACCpublic],
	    name="<init>",
	    msig = ([]: jtype list, NONE: jtype option),
	    attrs = [CODE{stack=1, locals=1, 
			  code=[Jaload (Localvar.fromInt 0),
				Jinvokespecial (defInit), Jreturn],
			  (* Have to be careful here;  
			     what if user has defined own initialiser???? *)
			  hdls=[]: exn_hdl list, 
			  attrs=[]: attribute list }] }
in
    if List.exists (fn d => (#name(d) = "<init>")) l then l else init::l
end

(* The type annotations above are there to prevent warnings 
   about value polymorphism *)



(*************************************)
(*          Main function            *)  
(*************************************)

fun compile file outputdir generateThy = 
(* outputdir = name of directory in which to write class files;  must end with "/" *)
let
    fun basename f = List.last (String.tokens (fn c => ( c = #"/")) f);

    val javaObject = SOME (class{pkgs=["java","lang"], name="Object"})

    val CDEF(flags, cname, fdefs, mdefs) = MainParser.parse file;

    val _ = if generateThy then 
		let val thy = TextIO.openOut (outputdir ^ cname ^ ".thy");
                    val () = printGrailPROG thy (CDEF(flags, cname, fdefs, mdefs))
		    val () = TextIO.closeOut thy
		in () end
	    else ();

    val cf = map (fn FDEF(flags,ty,name) 
		  => {flags=flags, name=name, 
		      ty=tyToJty(ty), attrs=[]}) fdefs
	
    (* What about initialising constant values? *)

    val cm = map (compile_mdef cname) mdefs;
    
    val cl: class_decl = { flags = ACCsuper::flags,
			  this = class{pkgs=[], name=cname},
			  super = javaObject, 
			  ifcs = [],
			  fdecls = cf,
			  mdecls = add_initialiser cm,
			  attrs=[SRCFILE(basename file)]};

(* WARNING:  This provides a default initialiser (the one 
for java.lang.Object) for every method.  There's as yet no 
means for the user to specify an initialiser *)


    val cp = Constpool.create();

    val os = BinIO.openOut (outputdir ^ cname ^ ".class")

    fun emitword w = BinIO.output1 (os, w);
    val () = Classfile.emit emitword cp cl
    val () = BinIO.closeOut os

in 
    cname
end;
