open GrailAbsyn;
open ToyClassTable;

datatype 'a ToyClassDef = ToyCDEF of cname list * mname list * ifldname list * rfldname list * 
                                 funame list * exprname list * iname list * rname list *
                                 failname list * ('a ToyExpression) list


and 'a ToyExpression = 
       Null_
     | Int_ of int
     | IVar of iname
     | RVar of rname
     | ToyPrimOp of ToyBinOp * iname * iname
     | ToyRPrimOp of ToyRBinOp * rname * rname
     | New_ of cname * (ifldname * iname) list * (rfldname * rname) list
     | GetFi of rname * ifldname
     | GetFr of rname * rfldname
     | PutFi of rname * ifldname * iname
     | PutFr of rname * rfldname * rname
     | Invoke of rname * mname * rname
     | InvokeStatic of cname * mname * rname
     | Leti of iname * 'a ToyExpression * 'a ToyExpression
     | Letr of rname * 'a ToyExpression * 'a ToyExpression
     | Ifg of iname * 'a ToyExpression * 'a ToyExpression
     | Call of funame
     | Ill_Expr

and ToyBinOp = ADDopToy | SUBopToy | MULopToy | DIVopToy | MODopToy | 
                EQtestToy | LtestToy | LEtestToy| GtestToy | GEtestToy | NEtestToy 

and ToyRBinOp =  REQtestToy | RUndeftestToy
 
and ToyTy = IntTy | RefTy | FailTy

withtype iname = string
        and rname = string  
        and cname = string
        and ifldname = string
        and rfldname = string
        and mname = string  
        and funame = string
        and exprname = string
        and failname = string;


val prseparator =  ref 0;
val testcounter = ref 0;


fun testToToy EQtest = EQtestToy
  | testToToy NEtest = NEtestToy
  | testToToy Ltest  = LtestToy
  | testToToy LEtest = LEtestToy
  | testToToy Gtest  = GtestToy
  | testToToy GEtest = GEtestToy

fun rTestToToy EQtest = REQtestToy
   | rTestToToy _ = RUndeftestToy

fun binOpToToy ADDop = ADDopToy
  | binOpToToy SUBop = SUBopToy
  | binOpToToy MULop = MULopToy
  | binOpToToy DIVop = DIVopToy
  | binOpToToy MODop = MODopToy

fun toyBinOpToString EQtestToy = "(%  x y . if y=x then (1::int) else (0::int))"
  | toyBinOpToString NEtestToy = "(%  x y . if y=x then (0::int) else (1::int))"
  | toyBinOpToString LtestToy  = "(%  x y . if x<y then (1::int) else (0::int))"
  | toyBinOpToString LEtestToy = "(%  x y . if y<x then (0::int) else (1::int))"
  | toyBinOpToString GtestToy  = "(%  x y . if y<x then (1::int) else (0::int))"
  | toyBinOpToString GEtestToy = "(%  x y . if x<y then (0::int) else (1::int))"
  | toyBinOpToString ADDopToy = "(%  x y . (x + y))"
  | toyBinOpToString SUBopToy = "(%  x y . (x - y))"
  | toyBinOpToString MULopToy = "(%  x y . (x * y))"
  | toyBinOpToString DIVopToy = "(%  x y . (x div y))"
  | toyBinOpToString MODopToy = "(%  x y . (x mod y))"

fun toyRBinOpToString REQtestToy = "(%  x y . if y=x then (1::int) else (0::int))"
  | toyRBinOpToString _ = "Wrong Refernce Test"

fun  typeFromGrailTy INTty = IntTy
   | typeFromGrailTy (REFty _) = RefTy
   | typeFromGrailTy _  = FailTy

fun typeFromGrailRTy rty =
    case rty of 
        NONE => FailTy
      | SOME INTty => IntTy     
      | SOME (REFty _ ) => RefTy
      | SOME  _    => FailTy



fun splitBy x xs = String.tokens (fn c => c=x) xs;

fun substrAfterToken x xs =
    List.last (splitBy x xs);

fun substrBeforeToken x xs =
    List.hd (splitBy x xs);


fun pointToPrime x =
    if (x= #".")
    then  #"'"
    else x

fun addElemToList elem l = 
     if not (List.exists (fn x => x=elem) l)
     then elem::l
     else l

fun  addLists [] l = l
   | addLists (h::t) l =
     addElemToList h (addLists t l)


fun mergeVectorsLists (l1, l2, l3, l4, l5, l6, l7, l8) (m1, m2, m3, m4, m5, m6, m7, m8) =
    (addLists l1 m1,
     addLists l2 m2,
     addLists l3 m3,
     addLists l4 m4,
     addLists l5 m5,
     addLists l6 m6,
     addLists l7 m7,
     addLists l8 m8)

fun initList [] = []
  | initList (h::[])= []
  | initList (h1::(h2::t)) = h1 :: (initList (h2::t)) 

fun prefixToList str [] = []
  | prefixToList str (h::t) =
    ((str^h) :: (prefixToList str t))


fun printGrailPROG thy (CDEF(flags, cname, fdefs, mdefs))=
    printToyPROG thy cname (translateGrailPROG (CDEF(flags, cname, fdefs, mdefs)))

and translateGrailPROG (CDEF(flags, cname, fdefs, mdefs))=
    let val (iflds, rflds, flnames) = getThisFldNames cname fdefs
        val (ToyCDEF(cnames, mnames, ifldnames, rfldnames, funames, 
                     exprnames, inames, rnames, failnames, exprs)) =
                getListsFromMthds cname mdefs 
    in  ToyCDEF(addElemToList cname cnames, 
                mnames, 
                addLists iflds ifldnames, 
                addLists rflds rfldnames,
                funames, exprnames, inames, rnames, 
                addLists flnames failnames, 
                exprs)
   end

and getListsFromMthds cname [] =  ToyCDEF([], [], [], [], [], [], [], [], [], [])
  | getListsFromMthds cname (h::t) = 
    let val (ToyCDEF(hcnames, hmnames, hifldnames, hrfldnames, hfunames, 
               hexprnames, hinames, hrnames, hfailnames, hexprs)) = 
           getListsFromMethod cname h
        val ToyCDEF(tcnames, tmnames, tifldnames, trfldnames, tfunames, 
                    texprnames, tinames, trnames, tfailnames, texprs) =
            getListsFromMthds cname t
    in ToyCDEF(
              (* stuff that may be shared by the methods inside the class *)
               addLists hcnames tcnames,        
               addLists hmnames tmnames,

              (* Filed names are already prefixed by the name of a holding class *)
               addLists hifldnames tifldnames,        
               addLists hrfldnames trfldnames,

              (* The names below are prefixed by class and method names  *)
              (* For instance, clname'metname'myname  *)
               hfunames @ tfunames,        
               hexprnames @ texprnames,
               hinames @ tinames,        
               hrnames @ trnames,
               hfailnames @ tfailnames,        
               hexprs @ texprs)
                             
    end

and getThisFldNames cname [] = ([], [], [])
| getThisFldNames cname ((FDEF(flags, ty, name))::t) =
  let val (ifldnames, rfldnames, failnames) = getThisFldNames cname t
  in case ty of
         INTty     => (addElemToList (cname^"'"^name) ifldnames, rfldnames, failnames)
       | (REFty _) => (ifldnames, addElemToList (cname^"'"^name) rfldnames, failnames)
       | _         => (ifldnames, rfldnames, addElemToList (cname^"'"^name) failnames)
  end

and getListsFromMethod cname (MDEF(flags, rty, name, params, mbody)) =
    (
    testcounter:=0;
    let 
        (* The first run of the compiler : getting all the names,
           not prefixed so far *)
        val (cnames, mnames, ifldnames, rfldnames, funames, inames, rnames, failnames) =
            getNamesFromMethod params mbody
    in (testcounter:=0; 
        let 
       (* the second run of the compiler : translation *)
        val fullmname = cname^"'"^name
        val (exprnames, exprs) = getExprsFromMethod  inames rnames (fullmname^"'") mbody
        in ToyCDEF(cnames, 
                   addElemToList fullmname mnames, 
                   ifldnames, rfldnames, 
                   prefixToList  (fullmname^"'") funames, 
                   prefixToList  (fullmname^"'") exprnames, 
                   prefixToList  (fullmname^"'") inames, 
                   prefixToList  (fullmname^"'") rnames, 
                   prefixToList  (fullmname^"'") failnames, 
                   exprs)
        end
       )
    end
    )

and getNamesFromMethod params (MBODY(letdecs, fundecs, result)) =
    let val (inames, rnames, failnames) = getNamesFromParams params 
    in  getNamesFromMBody inames rnames failnames letdecs fundecs result
    end

and getNamesFromParams [] = ([], [], [])
|   getNamesFromParams ((ty,va)::t) =
    let val (inames, rnames, failnames)  = getNamesFromParams t
    in case ty of
         INTty  => (va :: inames, rnames, failnames)
       | (REFty _) => (inames, va :: rnames, failnames)
       | _         => (inames, rnames, va :: failnames)
    end

and getNamesFromMBody inames rnames failnames letdecs fundecs result =
    let val (fnames, fletdecs, fresults) = splitFundecs fundecs
    in let val (cnameslet, mnameslet, ifldnameslet, rfldnameslet, funameslet, inameslet, rnameslet, failnameslet) = 
                  getNamesFromLetdecs inames rnames (letdecs @ fletdecs)
       in let val resultsnames = getNamesFromResults inameslet rnameslet (fresults @ (result :: []))
          in mergeVectorsLists 

                (cnameslet, mnameslet, ifldnameslet, rfldnameslet, 
                addLists fnames funameslet, 
                inameslet, rnameslet, 
                addLists failnames failnameslet)

                resultsnames
          end
       end
   end

(* so far function args are ignored *)
(* but they must be declared somewhere in the method, 
so there is no need to scan the list of parameters *)

and splitFundecs [] = ([], [], [])
|  splitFundecs ((FDEC(fname, args, FUNbody(letdecs, result))) :: t) = 
   let val (tfnames, tletdecs, tresults) = splitFundecs t
   in (fname::tfnames, letdecs @ tletdecs, result::tresults)
   end

and getNamesFromLetdecs inames rnames [] = 
              ([], [], [], [], [], inames, rnames, [])
 |  getNamesFromLetdecs inames rnames (h::t) =
    let val (cn, mn, ifldn, rfldn, fnn, inn, rn, failn) = getNamesFromLetdecs inames rnames (initList (h::t))
    in  let val vl = getNamesFromLetDec inn rn (List.last (h::t))  
        in mergeVectorsLists (cn, mn, ifldn, rfldn, fnn, inn, rn, failn) vl
        end
    end
    

and  getNamesFromLetDec inames rnames  (VALdec(id, p)) =
     let val ty = typeFromPrimOp inames rnames p
         val (cn, mn, ifldn, rfldn, fnn, inn, rn, failn) = getNamesFromPrimOp inames rnames p
     in case ty of
        IntTy => (cn, mn, ifldn, rfldn, fnn, addElemToList id inn, rn, failn)
        | RefTy => (cn, mn, ifldn, rfldn, fnn, inn, addElemToList id rn, failn)
        | _ => (cn, mn, ifldn, rfldn, fnn, inn, rn, addElemToList id failn)
     end  
 
(* Void dec can be only due to "putfield" *)
(* No new identifier appears, the names must be obtained from "putfield" arguments *)
|  getNamesFromLetDec inames rnames  (VOIDdec p) =
        getNamesFromPrimOp inames rnames p

and  getNamesFromPrimOp ideclared rdeclared (VALop p) = getNamesFromNonOperand ideclared rdeclared p
  | getNamesFromPrimOp ideclared rdeclared (BINop(b, v, w)) = 
    let val lv  = getNamesFromOperand ideclared rdeclared v
        val lw  = getNamesFromOperand ideclared rdeclared w
    in mergeVectorsLists lv lw
    end 

 
  | getNamesFromPrimOp ideclared rdeclared (NEWop(MDESC(rty, name, typs), l)) = 
    let val (inames, rnames, failnames) = getNamesFromListOperand ideclared rdeclared l
        val (iflds, rflds, failns)  = getFlds name
    in ([name], [], iflds, rflds, [], inames, rnames, failnames)
    end 

(* so far only one ref parameter is supported *)   
  | getNamesFromPrimOp ideclared rdeclared (INVOKEVIRTUALop(var, MDESC(rty, name, typs) , (VARval(vname)::[]))) = 
    if (List.exists (fn x => x=var) rdeclared)
    then if (List.exists (fn x => x=vname) rdeclared)
         then let val cname = (substrBeforeToken #"." name)
                  val mname = (String.map pointToPrime name)
                  (* the operands are already listed *)
               in (cname::[], mname::[], [], [], [], [], [] , [])
               end
         else ([], [], [], [], [], [], [] , vname :: [])
    else ([], [], [], [], [], [], [] , var::[])

  | getNamesFromPrimOp ideclared rdeclared (INVOKESTATICop(MDESC(rty, name, typs) , (VARval(vname)::[]))) =
    if (List.exists (fn x => x=vname) rdeclared)
     then let val cname = (substrBeforeToken #"." name)
              val mname = (String.map pointToPrime name)
               (* the operands are already listed *)
          in (cname::[], mname::[], [], [], [], [], [] , [])
          end
     else ([], [], [], [], [], [], [] , vname :: [])


  | getNamesFromPrimOp ideclared rdeclared (GETFIELDop(v, (FDESC(ty, name )))) = 
    if (List.exists (fn x => x=v) rdeclared)
    then let val cname = (substrBeforeToken #"." name)
             val fname = (String.map pointToPrime name)
         in   case ty of
         	INTty => (cname::[], [], fname::[], [], [], [], [] , [])
         	| (REFty _) => (cname::[], [], [], fname::[], [], [], [] , [])
         	| _ => (cname::[], [], [], [], [], [], [] , fname::[])
         end
    else ([], [], [], [], [], [], [] , v::[])

  | getNamesFromPrimOp ideclared rdeclared (PUTFIELDop(v, (FDESC(ty, name )), w)) = 
    if (List.exists (fn x => x=v) rdeclared)
    then let val lw= getNamesFromOperand ideclared rdeclared w
             val cname = (substrBeforeToken #"." name)
             val fname = (String.map pointToPrime name)
         in 
         case ty of
         INTty => mergeVectorsLists lw (cname::[], [], fname::[], [], [], [], [] , [])
         | (REFty _) => mergeVectorsLists lw (cname::[], [], [], fname::[], [], [], [] , [])
         | _ => mergeVectorsLists lw (cname::[], [], [], [], [], [], [] , fname::[])
         end
    else ([], [], [], [], [], [], [] , v::[])

  | getNamesFromPrimOp ideclared rdeclared   _ = ([], [], [], [], [], [], [] , [])


and getNamesFromListOperand ideclared rdeclared [] = ([], [], [])
   |getNamesFromListOperand ideclared rdeclared (h::t) =
    let val (cn, mn, fnn, ifldn, rfldn, inn, rn, failn) = getNamesFromOperand ideclared rdeclared h
        val (innt, rnt, failnt) = getNamesFromListOperand ideclared rdeclared t
    in (addLists inn innt, addLists rn rnt, addLists failn failnt)
    end

(* Values no taking part in binops are either constants (then now new name appears) *)
(* or variables, declared above. Therefore, no new names appear. *)
and getNamesFromNonOperand ideclared rdeclared p = ([], [], [], [], [], [], [] , [])


and getNamesFromOperand ideclared rdeclared (INTval iconst) = ([], [], [], [], [], ("c'"^(Int.toString iconst))::[], [] , [])
 |  getNamesFromOperand ideclared rdeclared (VARval vname) = 
    let val ty = typeFromVar ideclared rdeclared vname
    in if (ty=FailTy)
       then ([], [], [], [], [], [], [] , vname :: [])
       else (*the variable is already declared*)
            ([], [], [], [], [], [], [] , [])
    end
 |  getNamesFromOperand ideclared rdeclared (NULLval _) = ([], [], [], [], [], [], ["Null'FakeVar"] , [])
 |  getNamesFromOperand ideclared rdeclared _  = ([], [], [], [], [], [], [] , [])

and getNamesFromResults ideclared rdeclared [] = ([], [], [], [], [], [], [] , [])
 | getNamesFromResults  ideclared rdeclared (h::t) =
   mergeVectorsLists (getNamesFromResult ideclared rdeclared h) (getNamesFromResults ideclared rdeclared t)


and getNamesFromResult ideclared rdeclared (PRIMres p) =   getNamesFromPrimRes ideclared rdeclared p
 |  getNamesFromResult ideclared rdeclared (CHOICEres (v, tst, w, p1, p2)) =
    (testcounter := 1+(!testcounter); 
    let val newtest = "q'"^(Int.toString(!testcounter))
        val lv = getNamesFromOperand ideclared rdeclared v
        val lw = getNamesFromOperand ideclared rdeclared w
        val lp1 =  getNamesFromPrimRes ideclared rdeclared p1
        val lp2 =  getNamesFromPrimRes ideclared rdeclared p2     
    in mergeVectorsLists ([], [], [], [], [], newtest::[], [], [])
            (mergeVectorsLists lv (mergeVectorsLists lw (mergeVectorsLists lp1 lp2))) 
    end
    )

 

and getNamesFromPrimRes ideclared rdeclared (OPres p) = getNamesFromPrimOp ideclared rdeclared p
  | getNamesFromPrimRes ideclared rdeclared (VOIDres) = ([], [], [], [], [], [], [] , [])
  | getNamesFromPrimRes ideclared rdeclared (FUNres (name, args)) = ([], [], [], [], name :: [], [], [], [])


and getExprsFromMethod inames rnames prefix (MBODY(letdecs, fundecs, result)) = 
       getExprsFromNamedLetDecs inames rnames prefix (("", letdecs, result) ::  (makeNamedLetDecsList fundecs))


and makeNamedLetDecsList [] = []
 |  makeNamedLetDecsList ((FDEC(fname, args, (FUNbody(letdecs, result)))):: t) =
        ((fname, letdecs, result) :: (makeNamedLetDecsList t))

and  getExprsFromNamedLetDecs inames rnames prefix []  = ([], [])
  |  getExprsFromNamedLetDecs inames rnames prefix ((name, letdecs, result) :: t) = 
     let val newexpr = translateLetExprConsts inames inames rnames prefix letdecs result
         val (exprnamest, exprst) = getExprsFromNamedLetDecs inames rnames prefix t
     in ((name^"Body")::exprnamest, newexpr :: exprst)
     end

(* We start let-decs with declaration of all the integer constants,
  appearing in the code ! *)
(* The first argument is the list of integer names,
   which we check on the presence of constants ... *) 
and translateLetExprConsts [] inames rnames prefix letdecs result = 
       translateLetExprNull inames rnames prefix letdecs result
 | translateLetExprConsts (h::t) inames rnames prefix letdecs result =
   if (String.size h >2)
   then if (String.substring(h, 0, 2)="c'")
    	then 	let val x = Int.fromString (String.substring(h, 2, (String.size h)-2))
         	in  case x of 
               		(SOME i) => Leti(prefix^h, Int_(i), translateLetExprConsts t inames rnames prefix letdecs result) 
              		| _ => Ill_Expr
         	end
        else translateLetExprConsts t inames rnames prefix letdecs result
    else translateLetExprConsts t inames rnames prefix letdecs result

(* Now we need to check if Null'FakeVar appears in the list of rnames *)
(* If yes, then we need to declare it *)
and translateLetExprNull inames rnames prefix letdecs result = 
if (List.exists (fn x => x="Null'FakeVar") rnames)
     then  Letr(prefix^"Null'FakeVar", 
                         Null_, translateLetExpr inames rnames prefix letdecs result)
     else (translateLetExpr inames rnames prefix letdecs result) 

and translateLetExpr inames rnames prefix [] result =translateResult inames rnames prefix result
  | translateLetExpr inames rnames prefix ((VALdec(id, p))::t) result = 
       translateFlLetExpr inames rnames prefix id p t result
  | translateLetExpr inames rnames prefix ((VOIDdec p)::t) result = 
    case p of  
       (PUTFIELDop(v,(FDESC(ty, name )) , w)) => 
               translateFlLetExpr  inames rnames  prefix (translateOperand inames rnames "" "" w) p t result 
       | _ => translateFlLetExpr  inames rnames  prefix "fail'name" p t result   
      

and translateFlLetExpr inames rnames prefix id p t result =
    	if (List.exists (fn x => x=id) inames)
    	then  Leti(prefix^id, translatePrimOp inames rnames prefix p, translateLetExpr inames rnames prefix t result)
    	else if (List.exists (fn x => x=id) rnames)
 	     then  Letr(prefix^id, translatePrimOp inames rnames prefix p, translateLetExpr inames rnames prefix t result)
             else Ill_Expr


and translateResult ideclared rdeclared prefix (PRIMres p) = translatePrimRes ideclared rdeclared prefix p
  | translateResult ideclared rdeclared prefix (CHOICEres (v, tst, w, p1, p2)) =
    (
    testcounter := 1+(!testcounter);
    let val newtest = prefix^"q'"^(Int.toString(!testcounter))
        val vname = translateOperand ideclared rdeclared "" prefix v
        val wname = translateOperand ideclared rdeclared "" prefix w
        val translatedif = 
             Ifg(newtest, translatePrimRes ideclared rdeclared prefix p1 , translatePrimRes ideclared rdeclared prefix p2)
    in case v of 
        (NULLval _) => Leti(newtest, ToyRPrimOp(rTestToToy(tst), vname, wname), translatedif)  
      | _  => case w of 
              (NULLval _) => Leti(newtest, ToyRPrimOp(rTestToToy(tst), vname, wname), translatedif)
            | _ => Leti(newtest, ToyPrimOp(testToToy(tst), vname, wname), translatedif)                                                
    end
    )

and translatePrimRes ideclared rdeclared prefix (OPres p) = translatePrimOp ideclared rdeclared prefix p 
  | translatePrimRes ideclared rdeclared prefix (VOIDres) = Ill_Expr
  | translatePrimRes ideclared rdeclared prefix (FUNres (name, args)) = Call (prefix^name)

and typeFromPrimOp ideclared rdeclared (VALop (INTval iconst)) = IntTy
  | typeFromPrimOp ideclared rdeclared (VALop (VARval vname))= typeFromVar ideclared rdeclared vname
  | typeFromPrimOp ideclared rdeclared (VALop (NULLval _ ))=  RefTy
  | typeFromPrimOp ideclared rdeclared (BINop(b, v, w)) = IntTy
  | typeFromPrimOp ideclared rdeclared (NEWop(m,v)) = RefTy
  | typeFromPrimOp ideclared rdeclared (INVOKEVIRTUALop(var, MDESC(rty, name, typs) , params)) = typeFromGrailRTy rty
  | typeFromPrimOp ideclared rdeclared (INVOKESTATICop(MDESC(rty, name, typs) , params)) = typeFromGrailRTy rty
  | typeFromPrimOp ideclared rdeclared (GETFIELDop(v, (FDESC(ty, name )))) = typeFromGrailTy ty
  | typeFromPrimOp ideclared rdeclared (PUTFIELDop(v, (FDESC(ty, name )), w)) = typeFromGrailTy ty
  | typeFromPrimOp ideclared rdeclared   _ = FailTy

and typeFromVar ideclared rdeclared vname =
     if (List.exists (fn x => x=vname) ideclared)
     then IntTy
     else if (List.exists (fn x => x=vname) rdeclared)
          then RefTy
          else FailTy

and translatePrimOp ideclared rdeclared prefix (VALop v) = translateNonOperand ideclared rdeclared prefix v
  | translatePrimOp ideclared rdeclared prefix (BINop(b, v, w)) =
    ToyPrimOp(binOpToToy(b),
              translateOperand ideclared rdeclared "iname" prefix v,
              translateOperand ideclared rdeclared "iname" prefix w)


  | translatePrimOp  ideclared rdeclared prefix (NEWop(MDESC(rty, name, typs),l)) = 
      let val (iflds, rflds, failns)  = getFlds name
          val (inames, rnames, failnames) = splitListNames ideclared rdeclared prefix typs l 
      in  New_(name, ListPair.zip(iflds, inames), ListPair.zip(rflds, rnames))
      end

(* so far only one refernce parameter is supported *)
  | translatePrimOp ideclared rdeclared prefix (INVOKEVIRTUALop(var, MDESC(rty, name, typs) , (h :: []))) = 
        Invoke(prefix^var, String.map pointToPrime name, translateOperand ideclared rdeclared "rname" prefix h)

  | translatePrimOp ideclared rdeclared prefix (INVOKESTATICop(MDESC(rty, name, typs) , (h :: []))) = 
    let val cname = (substrBeforeToken #"." name)
    in InvokeStatic(cname, String.map pointToPrime name, translateOperand ideclared rdeclared "rname" prefix h)
    end

      
  | translatePrimOp ideclared rdeclared prefix (GETFIELDop(v, (FDESC(ty, name )))) = 
    (case ty of
             INTty  => GetFi(prefix^v, (String.map pointToPrime (name))) 
           | (REFty _) => GetFr(prefix^v, (String.map pointToPrime (name)))
           | _     => Ill_Expr
      )

  | translatePrimOp ideclared rdeclared prefix (PUTFIELDop(v,(FDESC(ty, name )) , w)) = 
    let val wname = translateOperand ideclared rdeclared "" prefix w 
    in case ty of
        INTty       => PutFi(prefix^v, String.map pointToPrime name, wname)
       | (REFty _ ) => PutFr(prefix^v, String.map pointToPrime name, wname)
       | _  => Ill_Expr
    end 
    

  | translatePrimOp ideclared rdeclared prefix   _ = Ill_Expr 

(* A non-operand value must be an expression *)
and translateNonOperand ideclared rdeclared prefix (INTval iconst) = Int_ iconst
  | translateNonOperand ideclared rdeclared prefix (VARval vname) = 
       let val ty = typeFromVar ideclared rdeclared vname
       in case ty of 
          IntTy => IVar (prefix^vname)
        | RefTy => RVar (prefix^vname)
        | _     => Ill_Expr
       end
   | translateNonOperand ideclared rdeclared prefix (NULLval _ )  = Null_
   | translateNonOperand ideclared rdeclared prefix  _            = Ill_Expr

(* an operand value must be a name *)
and translateOperand  ideclared rdeclared typ prefix (INTval iconst)  = 
    (prefix^"c'"^(Int.toString iconst))

| translateOperand ideclared rdeclared   typ prefix (VARval vname) = 
  if (typ="") 
  then prefix^vname 
  else let val ty = typeFromVar ideclared rdeclared vname
       in case ty of 
          	IntTy => if (typ="iname")
                   then  prefix^vname 
                   else (prefix^"fail'"^vname)
        	| RefTy => if (typ="rname")
                   then prefix^vname 
                   else (prefix^"fail'"^vname)
        	| _ => (prefix^"fail'"^vname) 
      end
| translateOperand ideclared rdeclared typ prefix (NULLval _ )  = prefix^"Null'FakeVar"
| translateOperand ideclared rdeclared typ prefix  _ = prefix^"fail'operand" 


and splitListNames ideclared rdeclared prefix [] [] = ([], [], [])
  | splitListNames ideclared rdeclared prefix (htype::ttype) (hvalue::tvalue) =
    let val (inamest, rnamest, failnamest) = splitListNames ideclared rdeclared prefix ttype tvalue
        val newname = translateOperand  ideclared rdeclared "" prefix hvalue
    in case htype of 
         INTty => (newname ::inamest, rnamest, failnamest)
         | (REFty _) => (inamest, newname :: rnamest, failnamest)
         | _ => (inamest, rnamest,  newname :: failnamest)
    end
  | splitListNames ideclared rdeclared prefix _ _ = ([], [], [])

(* ----------------------------------------------------------------------- *)
(* ------ PRINTING TO ISABELLE THEORY ------------------------------------ *)


and printToyPROG  thy toycname (ToyCDEF(cnames, mnames, ifldnames, rfldnames, funames, 
                               exprnames, inames, rnames, failnames, exprs)) =
    (
    PrlF thy "(* This file was automatically generated by gdf *)\n";
    PrF thy "theory ";  
    PrF thy toycname;           
    PrlF thy " = ToyHLderived :";

    PrlF thy "\nconstdefs\nupperbound :: nat\n\"upperbound == 0\""; 

    PrlF thy ("\nlocale "^toycname^"_example = \n"); 
    PrF thy "fixes  ";
    prseparator := 0;

    printNames thy "cname" cnames;
    printNames thy "\"'a expr\"" exprnames;
    printNames thy "mname" mnames;
    printNames thy "funame" funames;
    printNames thy "ifldname" ifldnames;
    printNames thy "rfldname" rfldnames;
    printNames thy "iname" inames;
    printNames thy "rname" rnames;

    if not (List.null failnames)
    then PrlF thy "The following names caused type checking error:"
    else ();
    printNames thy "failname" failnames;
   
    PrlF thy ("\ndefines");
    printExps thy exprnames exprs; 

    PrF thy "\n\nassumes";
    prseparator := 0;
    printMap  thy "m" mnames;
    printMap  thy "f" funames;

    printLemma thy toycname mnames rnames;

    PrlF thy "\n\nend"
   )

and printNames thy ty [] = ()
  | printNames thy ty (h::t) = 
    (if (!prseparator>0)
     then PrF thy "and    "
     else prseparator :=1; 
     PrlF thy (h^" :: "^ty);
     printNames thy ty t
    )

and printMap thy ty [] = ()
  | printMap thy ty (h::t) =
    let val () = printMapElement thy ty  h
    in printMap thy ty t
    end

and printMapElement thy ty elem =
   ( if (!prseparator > 0)
     then PrF thy "\n and   "
     else (prseparator:=1;
            PrF thy "\n       "
            );
     (case ty of 
          "f" => PrF thy (elem^"_fnbdy: \"funtable  "^elem^" = "^elem^"Body\"")
        | "m" => let val clname = substrBeforeToken #"'" elem
                 in PrF thy (elem^"_mtbdy: \"methtable "^clname^"  "^elem^" = "^elem^"Body\"")
                 end
        |  _  =>  PrlF thy "WRONG CALL OF PRINTMAP FUNCTION"
     )
    ) 

and printExps thy [] [] = ()
  | printExps thy (h1::t1) [] = PrlF thy "ILL FORMED EXPRESSION LIST"
  | printExps thy [] (h2::t2) = PrlF thy  "ILL FORMED EXPRESSION LIST"
  | printExps thy (h1::t1) (h2::t2) = 
    (
    PrlF thy ("\""^h1^" ==");
    printFlatExp thy h2;
    if (List.null t1)
    then PrlF thy ("\"")
    else PrF thy ("\n\"\nand    ");
    printExps thy t1 t2
    )


and printFlatExp thy Null_ = PrF thy "expr.Null"
  | printFlatExp thy (Int_(iconst)) =  PrF thy ("expr.Int "^(Int.toString(iconst)))
  | printFlatExp thy (IVar(name)) =  PrF thy ("IVar "^name)
  | printFlatExp thy (RVar(name)) =  PrF thy ("RVar "^name)
  | printFlatExp thy (ToyPrimOp(p, name1, name2)) =  
     PrF thy ("Primop "^(toyBinOpToString p)^" "^name1^" "^name2)
  | printFlatExp thy (ToyRPrimOp(p, name1, name2)) =  
     PrF thy ("RPrimop "^(toyRBinOpToString p)^" "^name1^" "^name2)
  | printFlatExp thy (New_(name, ipairlist, rpairlist)) =  
         PrF thy ("NEW <"^name^"> ( \n["^(pairListToString ipairlist)^"\n],\n["^(pairListToString rpairlist)^"\n]\n)")
  | printFlatExp thy (GetFi(name1, name2)) =  PrF thy ("GetFi "^name1^" "^name2)
  | printFlatExp thy (GetFr(name1, name2)) =  PrF thy ("GetFr "^name1^" "^name2)
  | printFlatExp thy (PutFi(name1, name2, name3)) =  PrF thy ("PutFi "^name1^" "^name2^" "^name3)
  | printFlatExp thy (PutFr(name1, name2, name3)) =  PrF thy ("PutFr "^name1^" "^name2^" "^name3)
  | printFlatExp thy (Invoke(name1, name2, name3)) =   PrF thy ("Invoke "^name1^" "^name2^" "^name3)
  | printFlatExp thy (InvokeStatic(name1, name2, name3)) =   PrF thy ("InvokeStatic "^name1^" "^name2^" "^name3)
  | printFlatExp thy  (Leti(name, expr1, expr2)) = 
     (
      PrF thy ("LET "^name^" = ");
      printFlatExp thy expr1;
      PrlF thy "\nIN";
      printFlatExp thy  expr2;
      PrF thy "\nEND"
     )
  | printFlatExp thy  (Letr(name, expr1, expr2)) = 
     (
      PrF thy ("LET rf "^name^" = ");
      printFlatExp thy expr1;
      PrlF thy "\nIN";
      printFlatExp thy  expr2;
      PrF thy "\nEND"
     )
  | printFlatExp thy  (Ifg(name, expr1, expr2)) =
     (
      PrF thy ("\nIF "^name^" THEN ");
      printFlatExp thy expr1;
      PrF thy "\nELSE ";
      printFlatExp thy  expr2;
      PrF thy "\n"
     )
  | printFlatExp thy  (Call(name)) = 
    (PrF thy ("CALL "^name)
    )
  | printFlatExp thy  (IllExpr)     =  PrF thy "Ill-formed expression"
 
and pairListToString [] = ""
  | pairListToString ((fldname, varname)::t)=
    "\n("^fldname^", "^varname^")"^(pairListToString t)

and printLemma thy cname  (hmname :: tmanmes) (hrname :: trnames) =
   ( PrlF thy ("\n\n\nlemma (in "^cname^"_example)");
     PrlF thy "\"\\<Turnstile> {(z,s). card (fmap_dom (oheap s)) = z}";
     PrlF thy ("(InvokeStatic "^cname^" "^hmname^" "^hrname^")");
     PrlF thy "{(z,s,v). card (fmap_dom (oheap s)) = z + upperbound}\""
  )
 |  printLemma thy cname _ _ =
    PrlF thy "Wrong method invocation: cannot generate lemma"

