(* Wrapper for standalone version of gdf *)

exception gdfError of string;
fun usage() = TextIO.print "Usage: gdf [-c] [-t] [-d outputdir] file1 [file2 ...]\n\
\   -d dir  use dir as output directory\n\
\   -c      print class name\n\
\   -t      generate Isabelle .thy file as well\n"

fun doargs nil _ _ _ = usage ()
  | doargs ("-c"::t) dir _ generateThy = doargs t dir true generateThy
  | doargs ("-t"::t) dir printClassName _ = doargs t dir printClassName true
  | doargs ("-d"::t) _ printClassName generateThy = 
    let in case t of 
         nil => raise (gdfError "no directory specified for classfile output")
       | d::args' => 
	 	if not (Nonstdio.file_exists d) then raise (gdfError ("can't find directory " ^ d))
		else 
		   if not (FileSys.isDir d) then raise (gdfError (d ^ " is not a directory"))
		   else let 
        	        val dir = if (String.sub (d, (String.size d)-1)) = #"/" 
	 		          then d else d ^ "/"
			in
			    doargs args' dir printClassName generateThy
			end
    end
  | doargs args dir printClassName generateThy =
    let 
	val compileFile = 
	    if 
		printClassName 
            then (* tell the user the name of the class *)
		fn f => (TextIO.print ((Compile.compile f dir generateThy) ^ "\n"))
	    else (* print a (not very) helpful message *)
		fn f => let 
			   val _ = Compile.compile f dir generateThy
			in  
			   TextIO.print ("Compiled " ^ f ^ "\n")
			end
	in
	    app compileFile args
	end

val _ = doargs (CommandLine.arguments()) "" false false
	   handle gdfError s => TextIO.print ("Error: " ^ s ^"\n")
	   handle Compile.gdfError s => TextIO.print ("Error: " ^ s ^"\n");

