(* Classdecl.sml
 *
 * Peter Bertelsen
 * September 1997
 *)

local
    open Label Bytecode Jvmtype
in
    datatype access_flag =
	ACCpublic	(* field/method/class/interface *)
      | ACCprivate	(* field/method *)
      | ACCprotected	(* field/method *)
      | ACCstatic	(* field/method *)
      | ACCfinal	(* field/method/class *)
      | ACCvolatile	(* field *)
      | ACCtransient	(* field *)
      | ACCsynchronized	(* method *)
      | ACCnative	(* method *)
      | ACCabstract	(* method/class/interface *)
      | ACCsuper	(* class/interface *)
      | ACCinterface	(* interface *)

    type exn_hdl =   (* exception handler declaration *)
	{start: label,		(* start of handler scope *)
	 stop:  label,		(* end of handler scope *)
	 entry: label,		(* handler entry point *)
	 catch: jclass option}	(* class of handled exception; NONE = any *)

    datatype attribute =
        SRCFILE of string		 	(* source file name *)
      | CONSTVAL of jvm_const			(* constant value *)
      | CODE of {stack:  int,			(* max operand stack depth *)
		 locals: int,			(* max size of local vars *)
		 code:   jvm_instr list,	(* instruction sequence *)
		 hdls:   exn_hdl list,	 	(* exception handlers *)
		 attrs:  attribute list}
      | EXNS of jclass list		 	(* exception classes *)
      | LINENUM of {start: label,
		    line:  int} list
      | LOCALVAR of {from:  label,
		     thru:  label,
		     name:  string,
		     ty:    jtype,
		     index: Localvar.index} list
      | ATTR of {attr: string, info: Word8Vector.vector}

    type field_decl =	(* field declaration *)
	{flags:  access_flag list,	(* access flags *)
	 name:   string,		(* unqualified name *)
	 ty:     jtype,			(* field type *)
	 attrs:  attribute list}	(* field attributes *)

    type method_decl =	(* method declaration *)
	{flags:  access_flag list,	(* access flags *)
	 name:   string,		(* unqualified name *)
	 msig:   method_sig,		(* method signature *)
	 attrs:  attribute list}	(* method attributes *)

    type class_decl =	(* class/interface declaration *)
	{flags:  access_flag list,	(* access flags *)
	 this:   jclass,		(* this class/interface *)
	 super:  jclass option,		(* direct superclass *)
	 ifcs:   jclass list,		(* direct superinterfaces *)
	 fdecls: field_decl list,	(* field declarations *)
	 mdecls: method_decl list,	(* method declarations *)
	 attrs:  attribute list}
end (* local *)
