structure Bytecode = struct

(* Bytecode.sml
 *
 * Peter Bertelsen
 * December 1997
 *)


exception InternalError of string
exception Unimplemented of string

datatype jvm_const =
    Cnull
  | Cint    of Int32.int
  | Cfloat  of Real32.real
  | Clong   of Int64.int
  | Cdouble of Real64.real
  | Cstring of string

local
    open Label Localvar Jvmtype
in
    datatype class_ref =
	CLASS of jclass   (* a class or interface *)
      | ARRAY of jtype    (* an `array class' of the given element type *)

    type field_ref = {class : jclass,   (* not an array class *)
		      name  : string,
		      ty    : jtype}

    type method_ref = {class : jclass,  (* not an array class *)
		       name  : string,
		       msig  : method_sig}

    (* The type of JVM instructions, extended with labels *)
    datatype jvm_instr =
	Jlabel of label
      | Jsconst of string
      | Jaaload
      | Jaastore
      | Jaconst_null
      | Jaload of index
   (* | Jaload_0 | Jaload_1 | Jaload_2 | Jaload_3 *)
   (* | Janewarray of class_ref *)
   (* | Jareturn *)
      | Jarraylength
      | Jastore of index
   (* | Jastore_0 | Jastore_1 | Jastore_2 | Jastore_3 *)
      | Jathrow
      | Jbaload
      | Jbastore
   (* | Jbipush of int *)
      | Jcaload
      | Jcastore
      | Jcheckcast of class_ref
      | Jd2f
      | Jd2i
      | Jd2l
      | Jdadd
      | Jdaload
      | Jdastore
      | Jdcmpg | Jdcmpl
      | Jdconst of Real64.real
   (* | Jdconst_0 | Jdconst_1 *)
      | Jddiv
      | Jdload of index
   (* | Jdload_0 | Jdload_1 | Jdload_2 | Jdload_3 *)
      | Jdmul
      | Jdneg
      | Jdrem
   (* | Jdreturn *)
      | Jdstore of index
   (* | Jdstore_0 | Jdstore_1 | Jdstore_2 | Jdstore_3 *)
      | Jdsub
      | Jdup
      | Jdup_x1
      | Jdup_x2
      | Jdup2
      | Jdup2_x1
      | Jdup2_x2
      | Jf2d
      | Jf2i
      | Jf2l
      | Jfadd
      | Jfaload
      | Jfastore
      | Jfcmpg | Jfcmpl
      | Jfconst of real (*Real32.real*)
   (* | Jfconst_0 | Jfconst_1 | Jfconst_2 *)
      | Jfdiv
      | Jfload of index
   (* | Jfload_0 | Jfload_1 | Jfload_2 | Jfload_3 *)
      | Jfmul
      | Jfneg
      | Jfrem
   (* | Jfreturn *)
      | Jfstore of index
   (* | Jfstore_0 | Jfstore_1 | Jfstore_2 | Jfstore_3 *)
      | Jfsub
      | Jgetfield of field_ref
      | Jgetstatic of field_ref
      | Jgoto of label
   (* | Jgoto_w of label *)
      | Ji2b
      | Ji2c
      | Ji2d
      | Ji2f
      | Ji2l
      | Ji2s
      | Jiadd
      | Jiaload
      | Jiand
      | Jiastore
      | Jiconst of Int32.int
   (* | Jiconst_m1 | Jiconst_0 | Jiconst_1 | Jiconst_2 *)
   (* | Jiconst_3 | Jiconst_4 | Jiconst_5 *)
      | Jidiv
      | Jif_acmpeq of label | Jif_acmpne of label
      | Jif_icmpeq of label | Jif_icmpne of label | Jif_icmplt of label
      | Jif_icmpge of label | Jif_icmpgt of label | Jif_icmple of label
      | Jifeq of label | Jifne of label | Jiflt of label
      | Jifge of label | Jifgt of label | Jifle of label
      | Jifnonnull of label | Jifnull of label
      | Jiinc of {var   : index,
		  const : int}
      | Jiload of index
   (* | Jiload_0 | Jiload_1 | Jiload_2 | Jiload_3 *)
      | Jimul
      | Jineg
      | Jinstanceof of class_ref
      | Jinvokeinterface of method_ref
      | Jinvokespecial of method_ref
      | Jinvokestatic of method_ref
      | Jinvokevirtual of method_ref
      | Jior
      | Jirem
   (* | Jireturn *)
      | Jishl
      | Jishr
      | Jistore of index
   (* | Jistore_0 | Jistore_1 | Jistore_2 | Jistore_3 *)
      | Jisub
      | Jiushr
      | Jixor
      | Jjsr of label
   (* | Jjsr_w of label *)
      | Jl2d
      | Jl2f
      | Jl2i
      | Jladd
      | Jlaload
      | Jland
      | Jlastore
      | Jlcmp
      | Jlconst of Int64.int
   (* | Jlconst_0 | Jlconst_1 *)
   (* | Jldc of    jvm_const *)
   (* | Jldc_w of  jvm_const *)
   (* | Jldc2_w of jvm_const *)
      | Jldiv
      | Jlload of index
   (* | Jlload_0 | Jlload_1 | Jlload_2 | Jlload_3 *)
      | Jlmul
      | Jlneg
      | Jlookupswitch of {default : label,
			  cases   : (Int32.int * label) list}
      | Jlor
      | Jlrem
   (* | Jlreturn *)
      | Jlshl
      | Jlshr
      | Jlstore of index
   (* | Jlstore_0 | Jlstore_1 | Jlstore_2 | Jlstore_3 *)
      | Jlsub
      |  Jlushr
      | Jlxor
      | Jmonitorenter
      | Jmonitorexit
   (* | Jmultianewarray of {elem : jtype,
                            dim  : int} *)
      | Jnew of jclass
   (* | Jnewarray of jtype *)
      | Jnewarray of {elem : jtype,
		      dim  : int}
      | Jnop
      | Jpop
      | Jpop2
      | Jputfield of field_ref
      | Jputstatic of field_ref
      | Jret of index
      | Jreturn
      | Jsaload
      | Jsastore
   (* | Jsipush of int *)
      | Jswap
      | Jtableswitch of {default : label,
			 offset  : Int32.int,   (* key of first target *)
			 targets : label Vector.vector}
   (* | Jwide of jvm_instr *)

end (* local *)

(* NOTE: Some JVM instructions have been omitted in the jvm_instr
 * datatype.  For example, there is no need to distinguish aload 0
 * from aload_0 until the binary byte code is emitted; the emitter can
 * choose the most compact aload variant.  These are the translations
 * made by the bytecode emitter (Emitcode.emit):
 *
 *   Jsconst   ==> ldc or ldc_w
 *   Jaload    ==> aload_<n>, aload, or wide aload
 *   Jastore   ==> astore_<n>, astore, or wide astore
 *   Jdconst   ==> dconst_<d> or ldc2_w
 *   Jdload    ==> dload_<n>, dload, or wide dload
 *   Jdstore   ==> dstore_<n>, dstore, or wide dstore
 *   Jfconst   ==> fconst_<f>, ldc, or ldc_w
 *   Jfload    ==> fload_<n>, fload, or wide fload
 *   Jfstore   ==> fstore_<n>, fstore, or wide fstore
 *   Jgoto     ==> goto or goto_w
 *   Jiconst   ==> iconst_<i>, bipush, sipush, ldc, or ldc_w
 *   Jiinc     ==> iinc or wide iinc
 *   Jiload    ==> iload_<n>, iload, or wide iload
 *   Jistore   ==> istore_<n>, istore, or wide istore
 *   Jjsr      ==> jsr or jsr_w
 *   Jlconst   ==> lconst_<l> or ldc2_w
 *   Jlload    ==> lload_<n>, lload, or wide lload
 *   Jlstore   ==> lstore_<n>, lstore, or wide lstore
 *   Jnewarray ==> newarray, anewarray, or multianewarray
 *   Jret      ==> ret or wide ret
 *   Jreturn   ==> areturn, dreturn, freturn, ireturn, lreturn, or return
 *)


(* Utility functions *)
fun intConst   i = Jiconst (Int32.fromInt i)
fun wordConst  w = Jiconst (Int32.fromInt(Word.toIntX w))
fun charConst  c = Jiconst (Int32.fromInt(Char.ord c))
(* fun realConst  r = Jdconst (Real64.fromReal r)*)
fun realConst r = raise Fail "Not implemented yet: SML/NJ problems"
fun checkCast  c = Jcheckcast (CLASS c)
fun checkArray t = Jcheckcast (ARRAY t)

end
