(* Constpool.sig
 *
 * Peter Bertelsen
 * December 1997
 *)

exception InvalidEntry of string

type pool

eqtype index

datatype entry =
    CPutf8 of string
  | CPint of Int32.int
  | CPfloat of Word8Vector.vector
  | CPlong of Int64.int
  | CPdouble of Word8Vector.vector
  | CPclass of index
  | CPstring of index
  | CPfieldref of {class: index, nameType: index}
  | CPmethodref of {class: index, nameType: index}
  | CPimethodref of {class: index, nameType: index}
  | CPnametype of {name: index, desc: index}
  | CPunused  (* placeholder *)

val create : unit -> pool

val makeIndex  : int -> index
val indexValue : index -> int

val insUtf8       : pool -> string              -> index
val insInt        : pool -> Int32.int           -> index
val insFloat      : pool -> Real32.real         -> index
val insLong       : pool -> Int64.int           -> index
val insDouble     : pool -> Real64.real         -> index
val insClass      : pool -> Jvmtype.jclass      -> index
val insArrayClass : pool -> Jvmtype.jtype       -> index
val insString     : pool -> string              -> index
val insNameType   : pool -> {name : string,
			     desc : string}     -> index
val insFieldref   : pool -> Bytecode.field_ref  -> index
val insMethodref  : pool -> Bytecode.method_ref -> index
val insIMethodref : pool -> Bytecode.method_ref -> index
val insConst      : pool -> Bytecode.jvm_const  -> index

val lookup : pool -> index -> entry
val emit   : (Word8.word -> unit) -> pool -> unit
